/**
 * Created by janis dähne.
 */
import {LoginCredentialsForBackend, RegisterUserForBackend} from "../types/login";
import {UserDataFullBase} from "../types/userData";
import axios from 'axios'
import {RequestApiPrefix} from './constants'
import {ResponseCodes} from "../helpers/responseCodes";
import logger, {SeverityLevel} from '../helpers/logger'
import {ServerResponseWithData} from "../types/response";
import * as genericLayer from './genericLayer'


const fileName = 'loginLayer'
const controllerPrefix = "auth"

/**
 * login
 *
 * returns user data if all ok
 * returns null if all ok but wrong credentials
 *
 * reject if some error
 * @param {LoginCredentialsForBackend} credentials
 * @returns {Promise<UserDataFullBase>}
 */
export async function login(credentials: LoginCredentialsForBackend): Promise<UserDataFullBase> {

  logger.debug(fileName, 'login started', SeverityLevel.info)

  return new Promise<UserDataFullBase>((resolve, reject) => {
    axios.post(`${RequestApiPrefix}${controllerPrefix}/login`, credentials) //auth/login
      .then(res => {

        const data = res.data as ServerResponseWithData<UserDataFullBase>

        if (!data) {
          reject(null)
          return
        }

        if (data.code !== ResponseCodes.ok) {
          logger.debug(fileName, 'login error on success', SeverityLevel.error)
          resolve(null)
          return
        }

        logger.debug(fileName, 'login finished', SeverityLevel.info)

        resolve(data.data)

      })
      .catch(reason => {
        logger.debug(fileName, 'login error', SeverityLevel.error)

        reject(reason)
      })
  })
}

/**
 * logout
 * @returns {Promise<void>}
 */
export async function logout(): Promise<void> {

  logger.debug(fileName, 'logout started', SeverityLevel.info)

  return new Promise<void>((resolve, reject) => {

    axios.post(`${RequestApiPrefix}${controllerPrefix}/logout`)
      .then(res => {

        if (!res.data || res.data.code !== ResponseCodes.ok) {
          logger.debug(fileName, 'logout error on success (wrong response code)', SeverityLevel.error)
          reject(res)
          return
        }

        logger.debug(fileName, 'logout finished', SeverityLevel.info)
        resolve()

      })
      .catch(reason => {
        logger.debug(fileName, 'logout error', SeverityLevel.error)
        reject(reason)
      })
  })

}


export function getUserData(): Promise<UserDataFullBase> {
  return genericLayer.get<UserDataFullBase>(fileName, `auth/userdata`)
}


export async function registerUser(registerData: RegisterUserForBackend): Promise<void> {

  logger.debug(fileName, 'register started', SeverityLevel.info)

  return new Promise<void>((resolve, reject) => {

    axios.post(`${RequestApiPrefix}${controllerPrefix}/register`, registerData)
      .then(res => {

        if (!res.data || res.data.code !== ResponseCodes.ok) {
          logger.debug(fileName, 'register error on success (wrong response code)', SeverityLevel.error)
          reject(res)
          return
        }

        logger.debug(fileName, 'register finished', SeverityLevel.info)
        resolve()

      })
      .catch(reason => {
        logger.debug(fileName, 'register error', SeverityLevel.error)
        reject(reason)
      })
  })

}

