/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {Input, Button, Form, Header, Icon, Popup} from 'semantic-ui-react'
import {RootState} from "../../state/reducers/index";
import {
  setPassword, setUsername, reset as resetLoginData,
  setIsCheckingAlreadyLoggedIn, setIsLoggingIn, setIsRegisterFormDisplayed
} from "../../state/actions/loginActions";
import {getUserSettingsAndDataAsync} from "../../state/actions/settingsSite/userSettingsSiteCrudActions";
import {setUserData} from "../../state/actions/settingsSite/userSettingsSiteActions";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import MaterialInput from '../material/materialInput'

import * as loginLayer from '../../communicationLayer/loginLayer'
import {RouteComponentProps} from "react-router";
import {UserDataFullBase} from "../../types/userData";
import {notifyError, notifyWarning} from "../../helpers/notificationHelper";
import {OnKeyPressEvent} from "../../types/reactEvents";
import {delay} from "../../../debug";
import Spinner from "../helpers/Spinner";
import * as cryptoHelper from '../../helpers/cryptoHelper'
import {translate} from "../../../i18n/translationHelper";
import {TranslateIds} from "../../../i18n/ids";
import {FrontendSettingsManager} from "../../helpers/frontendSettingsManager";


//const css = require('./styles.styl');

export interface MyProps {
  //readonly caption?: string
}

const mapStateToProps = (rootState: RootState, props: MyProps) => {
  return {
    id: rootState.loginDataState.username,
    pw: rootState.loginDataState.password,
    isLogginIn: rootState.loginDataState.isLoggingIn,
    isCheckingAlreadyLoggedIn: rootState.loginDataState.isCheckingAlreadyLoggedIn,
  }
}

const mapDispatchToProps = (dispatch: Dispatch<any>) => bindActionCreators({
  setUsername,
  setPassword,
  getUserSettingsAndDataAsync,
  resetLoginData,

  setUserData,
  setIsCheckingAlreadyLoggedIn,
  setIsLoggingIn,
  setIsRegisterFormDisplayed
}, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;


export const formSpinnerOpacity = 0.3

class LoginForm extends React.Component<Props & RouteComponentProps<any>, any> {

  idChanged(e: React.FormEvent<HTMLInputElement>): void {
    this.props.setUsername(e.currentTarget.value)
  }

  pwChanged(e: React.FormEvent<HTMLInputElement>): void {
    this.props.setPassword(e.currentTarget.value)
  }

  componentDidMount(): void {


    // setTimeout(() => {
    //   this.loginClicked()
    // }, 200)
  }

  async loginClicked(): Promise<void> {

    this.props.setIsLoggingIn(true)

    // noinspection TsLint
    let userData: UserDataFullBase = null

    try {

      userData = await
        delay(
          loginLayer.login({
            token: this.props.id,
            password: cryptoHelper.hash(this.props.pw)
          })
        )

      if (userData === null) {
        notifyWarning(translate(TranslateIds.wrongUsernameOrPassword))
      } else {
        this.props.setUserData(userData)

        //clear login form
        this.props.resetLoginData()

        //clear frontend settings if the user changed...
        const frontendSettings = FrontendSettingsManager.getFrontendSettings()
        if (frontendSettings.lastLoggedInUserId !== userData.id) {
          FrontendSettingsManager.clear()
        }

        FrontendSettingsManager.setLastLoggedInUserId(userData.id)


      }
    } catch (err) {
      notifyError(translate(TranslateIds.serverError))
    }
    finally {
      this.props.setIsLoggingIn(false)
    }
  }

  render(): JSX.Element {

    return (
      <div className="frosted-glass fh">
        <div className="frosted-glass-inner absolute-sized-centered">

          <div className="corner-div fh">

            {
              this.props.isCheckingAlreadyLoggedIn &&
              <Spinner opacity={formSpinnerOpacity}/>
            }

            {
              this.props.isLogginIn &&
              <Spinner opacity={formSpinnerOpacity}/>
            }
            {
              (!this.props.isCheckingAlreadyLoggedIn && !this.props.isLogginIn) &&
              <div className="ui form frosted-glass-form">
                <div className="form-header">
                  <Header as='h1'>
                    {translate(TranslateIds.projectTitle)} - {translate(TranslateIds.login)}
                  </Header>
                </div>

                <Form.Field>
                  <div>
                    <div>
                      <Icon size="massive" name="user circle"/>
                    </div>
                  </div>
                </Form.Field>

                <Form.Field>
                  <label>
                    {
                      translate(TranslateIds.token)
                    }
                  </label>
                  <MaterialInput placeholder={`${translate(TranslateIds.token)}...`} value={this.props.id} onChange={(e) => this.idChanged(e)}
                                 onKeyPress={(e: OnKeyPressEvent) => {
                                   if (e.key === 'Enter') {
                                     this.loginClicked()
                                   }
                                 }}
                  />
                </Form.Field>

                <Form.Field className="relative-positioned">
                  <div className="password-forgotten">
                    <Popup
                      trigger={<a>
                        {translate(TranslateIds.forgotPassword)}
                      </a>}
                      content={
                        <div>
                    <span>
                      {
                        translate(TranslateIds.forgotPassword_text)
                      }
                    </span>
                        </div>
                      }
                      on='click'
                    />
                  </div>
                  <label>
                    {
                      translate(TranslateIds.password)
                    }
                  </label>
                  <MaterialInput type="password" placeholder={`${translate(TranslateIds.password)}...`} value={this.props.pw}
                                 onChange={e => this.pwChanged(e)}

                                 onKeyPress={(e: OnKeyPressEvent) => {
                                   if (e.key === 'Enter') {
                                     this.loginClicked()
                                   }
                                 }}
                  />


                </Form.Field>

                <div>
                  <Button color="blue"
                          onClick={e => this.loginClicked()}
                          content={translate(TranslateIds.signIn)}
                  >
                  </Button>
                  <Button
                    onClick={() => {
                      this.props.setIsRegisterFormDisplayed(true)
                    }}
                  >

                    <span>
                      {
                        translate(TranslateIds.register)
                      }
                    </span>
                    <Icon name="chevron right"/>
                  </Button>
                </div>

              </div>
            }
          </div>

        </div>

        <div className="absolute-sized-centered-below cookie-info-wrapper">
          <div>
            {
              translate(TranslateIds.cookie_info) + ' '
            }
            <a href='/copyright.html' target="_blank" className="terms-of-use-link">{
              translate(TranslateIds.cookie_info_link)
            }</a>
          </div>
        </div>
      </div>
    )
  }
}

export default connect(mapStateToProps, mapDispatchToProps)(LoginForm)

