/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../../../state/reducers/index";
import {TestProtocolItem, TestProtocolType} from "../../../../types/testResults";
import {List} from "semantic-ui-react";
import {Divider, Icon, Popup} from "semantic-ui-react";
import {leftIconPopupOffset} from "../../../../constants";
import * as JsDiff from "diff";
import {startsWith} from "../../../../helpers/stringHelper";
import {translate} from "../../../../../i18n/translationHelper";
import {TranslateIds} from "../../../../../i18n/ids";
import {HelpPopup} from "../../../helpers/helpPopup";


//const css = require('./styles.styl');

export interface MyProps {
  //readonly test: string
  readonly item: TestProtocolItem
  /**
   * the next item to display a diff if output mismatched (or null if it's the last)
   */
  readonly nextItem: TestProtocolItem | null

  /**
   * true: use js diff to display the diff between the expected output (unchanged) and the real output
   * (instead of the real output the diff is shown)
   * false: if output is wrong only mark it (e.g. red) but output real output
   */
  readonly useDiffIfOutputMismatched: boolean

  /**
   * true: replace all whitespaces by bullet signs
   */
  readonly useReplaceWhiteSpacesWithBullets: boolean
}

const mapStateToProps = (rootState: RootState, props: MyProps) => {
  return {
    //test0: rootState...
    //test: props.test
    item: props.item,
    nextItem: props.nextItem,
    useDiffIfOutputMismatched: props.useDiffIfOutputMismatched,
    useReplaceWhiteSpacesWithBullets: props.useReplaceWhiteSpacesWithBullets
  }
}

const mapDispatchToProps = (dispatch: Dispatch<any>) => bindActionCreators({
  //imported reducer funcs here

}, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;


// noinspection TsLint
/**
 * composes x funcs into one new function ... the functions are applied sequentially
 * @param {(line: string) => string} funcs
 * @returns {(line: string) => string}
 */
function compositeFuncs(...funcs: Array<(line: string) => string>): ((line: string) => string) {

  //function composition f(g(x))  in haskell we would write return (f . g)
  return (line) => {
    // noinspection TsLint
    let result = line
    for (const func of funcs) {
      result = func(result)
    }
    return result
  }
}

class consoleItem extends React.Component<Props, any> {
  render(): JSX.Element {

    const nextItemContent = this.props.nextItem ? this.props.nextItem.content : ''

    // noinspection TsLint
    let contentTransformer: (line: string) => string = line => line

    if (this.props.useReplaceWhiteSpacesWithBullets) {

      //contentTransformer = (line) => line.replace(/\s/g, "·") //matches any space...

      contentTransformer =
        compositeFuncs(
          //replace all spaces with ·
          (line) => line.replace(/ /g, "·"),
          //replace all tabs with ⇥ (mac tab sign)
          (line) => line.replace(/\t/g, "⇥")
        )
    }

    // noinspection TsLint
    return (
      <List.Item style={{padding: 0}}>
        {
          this.props.item.type === TestProtocolType.isProgramArguments &&
          <div className="flexed console-arguments-line">

            <HelpPopup className="mar-right-half"
                       defaultText={translate(TranslateIds.consoleArguments_helpText)}/>

            <div className="mar-right-half">
              {translate(TranslateIds.arguments_)}:
            </div>
            <div>
              <pre className="mar-null">
                {
                  contentTransformer(this.props.item.content)
                }
              </pre>
            </div>
          </div>
        }
        {
          this.props.item.type === TestProtocolType.isError &&
          <div className="error-colored">
            <pre className="mar-null console-item-pre">
              {
                //don't transform errors
                this.props.item.content
              }
            </pre>
          </div>
        }
        {
          this.props.item.type === TestProtocolType.unknown &&
          <div className="grey-colored">
            <pre className="mar-null console-item-pre">
              {
                `(${translate(TranslateIds.invalidTestResultLine)}):${contentTransformer(this.props.item.content)}`
              }
            </pre>
          </div>
        }
        {
          this.props.item.type === TestProtocolType.plain &&
          <div>
            <pre className="mar-null console-item-pre">
              {
                contentTransformer(this.props.item.content)
              }
            </pre>
          </div>
        }
        {
          //input
          this.props.item.type === TestProtocolType.isInput &&
          <div className="console-input-output-line">
            <div className="input-part light-grey-bg-colored">
              <div>
                <pre className="mar-null">
                {
                  contentTransformer(this.props.item.content)
                }
              </pre>
              </div>
            </div>
            <div className="program-part">

            </div>
            <div className="output-part">

            </div>
          </div>
        }
        {
          //output
          this.props.item.type === TestProtocolType.isOutput &&
          <div className="console-input-output-line">
            <div className="input-part">

            </div>
            <div className={this.props.item.hasOutputMismatched ? 'program-part-mismatch' : 'program-part'}>

            </div>
            <div className="output-part light-grey-bg-colored">

              {
                //all ok
                !this.props.item.hasOutputMismatched &&
                <div>
                  <pre className="mar-null">
                  {
                    contentTransformer(this.props.item.content)
                  }
                  </pre>
                </div>
              }

              {
                (this.props.item.hasOutputMismatched
                  &&
                  (
                    !this.props.useDiffIfOutputMismatched
                    || startsWith(this.props.item.content, "(regex):")
                  )
                ) &&
                <div className="error-colored">
                  <pre className="mar-null">
                  {
                    contentTransformer(this.props.item.content)
                  }
                  </pre>
                </div>
              }
              {
                //diff output, if we use diff output we cannot use regex because this doesn't make sense
                (this.props.item.hasOutputMismatched && this.props.useDiffIfOutputMismatched
                  && !startsWith(this.props.item.content, "(regex):")
                ) &&
                <div>
                  <pre className="mar-null">
                    {
                      JsDiff.diffChars(contentTransformer(nextItemContent),
                        contentTransformer(this.props.item.content)).map((part, index) => {
                        return (
                          <span key={index}>
                          {
                            (!part.added && !part.removed) &&
                            <span>{part.value}</span>
                          }
                            {
                              (part.added && !part.removed) &&
                              <span className="error-colored">{part.value}</span>
                            }
                            {
                              (!part.added && part.removed) &&
                              <span className="diff-blue-colored">{part.value}</span>
                            }
                       </span>

                        )
                      })
                    }
                  </pre>
                </div>
              }

            </div>
          </div>
        }
        {
          //expected output
          this.props.item.type === TestProtocolType.isExpectedOutput &&
          <div className="console-input-output-line">
            <div className="input-part">

            </div>

            {
              this.props.useDiffIfOutputMismatched &&
              <Popup
                trigger={
                  <div className="clickable program-part-expected-output">
                  </div>
                }
                content={
                  translate(TranslateIds.useDiffIfOutputMismatchedTrue_desc)
                }
                position="top left"
                offset={leftIconPopupOffset + 5}
                wide
              />
            }

            {
              !this.props.useDiffIfOutputMismatched &&
              <Popup
                trigger={
                  <div className="clickable program-part-expected-output">
                  </div>
                }
                content={translate(TranslateIds.useDiffIfOutputMismatchedFalse_desc)}
                position="top left"
                offset={leftIconPopupOffset + 5}
                wide
              />
            }


            <div className="expected-output-part">
              <div>
                <pre className="mar-null">
                  {
                    contentTransformer(this.props.item.content)
                  }
                </pre>
              </div>
            </div>
          </div>
        }

        {
          this.props.item.type === TestProtocolType.isEndOfTestLine &&
          <div>
            <Divider horizontal>
              {
                translate(TranslateIds.EOT)
              }
            </Divider>
          </div>
        }
      </List.Item>
    )
  }
}

export default connect(mapStateToProps, mapDispatchToProps)(consoleItem)