/**
 * Created by janis dähne.
 */
import {AwaitActions, MultiActions} from "../state/actions/types";
import {getMaxEntity} from "./mathHelper";
import {FileIdAssociation, SolutionFileDoExerciseFullBase} from "../types/exerciseSolution";
import {ExercisePreviewsFrontendExclusive} from "../components/sites/allAccessibleExercisesSite/openExercises/allOpenExercisesSite";
import {Moment} from "moment";
import moment = require("moment");
import {ReleaseDurationType} from "../types/ReleaseDurationType";
import {padLeft} from "./stringHelper";
import {errorDialog} from "./dialogHelper";
import {translate} from "../../i18n/translationHelper";
import {TranslateIds} from "../../i18n/ids";
import Logger from "./logger";
import {ReleaseStartType} from "../types/ReleaseStartType";
import {EditorTab, MultiTabEditorState} from "../components/codeEditors/multiTabCodeEditorProps";
import {delay, wait} from "../../debug";
import {ExerciseCodeTemplateFileFullBase} from "../types/exerciseEditor";


interface TabLike {
  readonly id: number
  readonly displayIndex: number
}

export const aceTextLanguage = 'text'

/**
 * this only sets the display indices DOES NOT CHANGE THE VISIBILITY
 * @param {TabLike} targetFile
 * @param {boolean} isDisplayed
 * @param {ReadonlyArray<TabLike>} templateFiles
 * @param {number} selectedFileId
 * @param {(newFiles: ReadonlyArray<TabLike>) => any} setFilesAction
 * @param {(newTabId: number) => any} setSelectedEditorTabIdAction
 * @returns {MultiActions}
 */
export function changeTabVisibility(targetFile: TabLike,
                                    isDisplayed: boolean,
                                    templateFiles: ReadonlyArray<TabLike>,
                                    selectedFileId: number,
                                    setFilesAction: (newFiles: ReadonlyArray<TabLike>) => void,
                                    setSelectedEditorTabIdAction: (newTabId: number) => void,): MultiActions {
  return dispatch => {

    //this would do nothing and it's a bug when we execute this because this would hide the file
    if (isDisplayed === true && targetFile.id === selectedFileId) return

    // noinspection TsLint
    let currDisplayIndex = 1
    // noinspection TsLint
    let currNotDisplayedIndex = -1

    const newTempFiles = templateFiles.map(file => {
      return file.id === targetFile.id
        ? {
          ...file,
          displayIndex: isDisplayed ? currDisplayIndex++ : currNotDisplayedIndex--
        }
        : {
          ...file,
          displayIndex: (file.displayIndex >= 0) ? currDisplayIndex++ : currNotDisplayedIndex--
        }
    })

    dispatch(setFilesAction(newTempFiles))

    //set new selected tab if necessary

    if (targetFile.id === selectedFileId) {
      //now selected the next file... or nothing

      const visibleFiles = newTempFiles.filter(p => p.displayIndex >= 0)

      if (visibleFiles.length === 0) {
        //no files is left open...
        dispatch(setSelectedEditorTabIdAction(-1))
      } else {

        const visibleFile = getMaxEntity(visibleFiles, p => p.displayIndex)

        dispatch(setSelectedEditorTabIdAction(visibleFile.id))
      }
    } else if (targetFile.id !== selectedFileId && isDisplayed === true) {

      //we clicked a file that is not the current active file ... make it the active file
      dispatch(setSelectedEditorTabIdAction(targetFile.id))

    }
  }
}

/**
 * checks if the options differs or not
 * @param {T} newSettings
 * @param {T} oldSettings
 * @returns {boolean}
 */
export function hasEditorSettingsChanged<T>(newSettings: T, oldSettings: T): boolean {

  if (!newSettings) return true
  if (!oldSettings) return true

  const oldKeys = Object.keys(oldSettings)
  const newKeys = Object.keys(newSettings)

  if (oldKeys.length !== newKeys.length) return true


  // noinspection TsLint
  for (const key in newSettings) {

    const oldKey = oldKeys.find(p => p === key)
    if (!oldKey) return false

    //check vals
    const newVal = newSettings[key]
    const oldVal = oldSettings[oldKey]

    if (newVal !== oldVal) return true

  }

  return false
}

export interface HasFileChangedMap {
  // noinspection TsLint
  /**
   * id: the file id,
   * bool: has changed?
   *
   * not readonly for performance
   */
  [id: string]: boolean
}

/**
 * updates (or creates) the hasFileChangedMap to display the has changed states in the editor
 * @param {EditorTab} changedFile
 * @param {ReadonlyArray<EditorTab>} allTabs
 * @param {HasFileChangedMap} hasFileChangedMap
 * @returns {HasFileChangedMap}
 */
export function updateHasFileChangedMap(changedFile: EditorTab, allTabs: ReadonlyArray<EditorTab>, hasFileChangedMap: HasFileChangedMap): HasFileChangedMap {

  // noinspection TsLint
  let copy: HasFileChangedMap = {}
  const currentFileKey = changedFile.id.toString()

  for (const tab of allTabs) {

    const key = tab.id.toString()
    // noinspection TsLint
    let hasChanged = hasFileChangedMap[key]

    if (hasChanged === undefined) {
      hasChanged = false
    }

    copy[key] = currentFileKey === key ? true : hasChanged
  }

  return copy

}


/**
 * DO NOT USE THIS !! only use this if really necessary
 *
 * this sets the frontend file ids to the backend file ids
 *
 * we cannot return a copy of the array else the editor undo won't work...
 * if we had only 1 version of the file in the backend then the id would stay
 * the same and this method would be not longer needed...
 *
 * @param {ReadonlyArray<SolutionFileDoExerciseFullBase>} frontendFiles
 * @param {ReadonlyArray<SolutionFileDoExerciseFullBase>} filesFromBackend
 * @param fileIdAssociationMap the map which stores which frontend id belongs to which backend saved file id
 * @returns {ReadonlyArray<SolutionFileDoExerciseFullBase>}
 */
export function updateFilesReference(frontendFiles: ReadonlyArray<SolutionFileDoExerciseFullBase>,
                                     filesFromBackend: ReadonlyArray<SolutionFileDoExerciseFullBase>,
                                     fileIdAssociationMap: ReadonlyArray<FileIdAssociation>): ReadonlyArray<SolutionFileDoExerciseFullBase> {


  //take the ids from the files from backend to the frontend files

  // noinspection TsLint
  for (let frontendFile of frontendFiles) {

    const fileIdTuple = fileIdAssociationMap.find(p => p.frontendFileId === frontendFile.id)

    if (fileIdTuple === undefined) {

      //we save the frontend version so if we create/delete a frontend file the backend should this too...
      //so normally this cannot happen

      errorDialog(
        translate(TranslateIds.error),
        translate(TranslateIds.couldNotFindSolutionFile)
      )
      continue
    }

    const backendFile: SolutionFileDoExerciseFullBase = filesFromBackend.find(p => p.id === fileIdTuple.backendFileId)

    if (!backendFile) {
      //TODO what on error?
      errorDialog(
        translate(TranslateIds.error),
        translate(TranslateIds.couldNotFindSolutionFile)
      )
      continue
    }

    frontendFile["i" + "d"] = backendFile.id
  }

  return frontendFiles
}


export function getNewSelectedTabId(frontendFiles: ReadonlyArray<SolutionFileDoExerciseFullBase>,
                                    filesFromBackend: ReadonlyArray<SolutionFileDoExerciseFullBase>,
                                    oldSelectedTabId: number): number {


  //take the ids from the files from backend to the frontend files
  // noinspection TsLint
  for (let frontendFile of frontendFiles) {

    const backendFile: SolutionFileDoExerciseFullBase = filesFromBackend.find(p => p.content === frontendFile.content)

    if (!backendFile) {
      //TODO what on error?
      errorDialog(
        translate(TranslateIds.error),
        translate(TranslateIds.couldNotFindSolutionFile)
      )
      continue
    }

    if (frontendFile.id === oldSelectedTabId) {
      return backendFile.id
    }
  }

  return -1
}


//recalculates time is up stuff
export function updateExercisePreviews(previews: ReadonlyArray<ExercisePreviewsFrontendExclusive>,
                                       clientServerTimeDiffInS: number): ReadonlyArray<ExercisePreviewsFrontendExclusive> {


  const serverNow: Moment = moment().add(clientServerTimeDiffInS, 'seconds')

  const copies = previews.map<ExercisePreviewsFrontendExclusive>(
    (preview: ExercisePreviewsFrontendExclusive) => {
      // noinspection TsLint
      const availableWorkingTimeInSeconds: number = preview.availableWorkingTimeInMinutes * 60


      if (preview.firstAccessedAt && !preview.firstAccessedAt.isValid()) {
        Logger.error(
          `firstAccessedAt was invalid: ${preview.firstAccessedAt.toString()}(editorHelper, updateExercisePreviews)`)
      }

      const isAllowedToSolve = _userIsAllowedToSolveExercisePreview(preview, serverNow)
      // noinspection TsLint
      let leftTimeString = ''
      // noinspection TsLint
      let leftTimeInS: number | null = null

      //set initial

      if (preview.isReleased && preview.hasLimitedWorkingTime === false && preview.releaseStartType === ReleaseStartType.Manual) {
        //we can solve this at any time
        const copy: ExercisePreviewsFrontendExclusive = {
          ...preview,
          isStillOpen: true,
          leftTimeInS
        }
        return copy
      }

      if (!preview.isReleased) {
        //we can solve this at any time
        const copy: ExercisePreviewsFrontendExclusive = {
          ...preview,
          isStillOpen: false,
          leftTimeInS
        }
        return copy
      }

      if (!preview.firstAccessedAt && preview.releaseDurationType === ReleaseDurationType.Infinitely) {
        //not yet accessed and no automatic end at --> user has full time left
        leftTimeString = `${preview.availableWorkingTimeInMinutes} ${translate(TranslateIds.minutes_abbr)}`
        // noinspection TsLint
        leftTimeInS = preview.availableWorkingTimeInMinutes * 60

        const copy: ExercisePreviewsFrontendExclusive = {
          ...preview,
          leftTimeString,
          leftTimeInS,
          isStillOpen: isAllowedToSolve,
        }
        return copy
      }

      if (preview.isReleased && preview.releaseStartType === ReleaseStartType.Automatic
        && preview.releaseDurationType === ReleaseDurationType.Infinitely && preview.hasLimitedWorkingTime === false) {
        //not yet accessed or automatic end at
        leftTimeString = `${preview.availableWorkingTimeInMinutes} ${translate(TranslateIds.minutes_abbr)}`
        // noinspection TsLint
        leftTimeInS = preview.availableWorkingTimeInMinutes * 60

        const copy: ExercisePreviewsFrontendExclusive = {
          ...preview,
          leftTimeString: '',
          leftTimeInS: null,
          isStillOpen: true,
        }
        return copy
      }

      // noinspection TsLint
      let diffBeforeReleaseEnd = Number.POSITIVE_INFINITY
      // noinspection TsLint
      let diffToWorkingTimeDeadlineInSeconds = Number.POSITIVE_INFINITY

      if (preview.firstAccessedAt && preview.hasLimitedWorkingTime) {
        // noinspection TsLint
        let diffInSeconds = serverNow.diff(preview.firstAccessedAt, 'seconds')
        diffToWorkingTimeDeadlineInSeconds = Math.floor(availableWorkingTimeInSeconds - diffInSeconds)
      }

      if (preview.automaticEndAt && preview.releaseStartType === ReleaseStartType.Automatic
        && preview.releaseDurationType === ReleaseDurationType.Limited) {
        diffBeforeReleaseEnd = preview.automaticEndAt.diff(serverNow, 'seconds')
      }

      if (preview.automaticEndAt && preview.releaseStartType === ReleaseStartType.Automatic
        && preview.releaseDurationType === ReleaseDurationType.Infinitely) {
        diffBeforeReleaseEnd = preview.automaticEndAt.diff(serverNow, 'seconds')
      }

      const minDeadlineInS = Math.min(diffToWorkingTimeDeadlineInSeconds, diffBeforeReleaseEnd)

      if (minDeadlineInS === Number.POSITIVE_INFINITY) {
        Logger.error("minDeadline was invalid (Number.POSITIVE_INFINITY) (editorHelper, updateExercisePreviews)")
        Logger.error(`firstAccessedAt: ${preview.firstAccessedAt.toString()} (editorHelper, updateExercisePreviews)`)
        Logger.error(`hasLimitedWorkingTime: ${preview.hasLimitedWorkingTime} (editorHelper, updateExercisePreviews)`)

        errorDialog(translate(TranslateIds.error),
          translate(TranslateIds.somethingWentWrongCalculatingLeftWorkingTime)
        )

        const copy: ExercisePreviewsFrontendExclusive = {
          ...preview,
          leftTimeString: translate(TranslateIds.error),
          leftTimeInS: null,
          isStillOpen: false,
        }
        return copy

      } else if (isNaN(minDeadlineInS)) {

        Logger.error("minDeadline was NaN (editorHelper, updateExercisePreviews)")
        Logger.error(`firstAccessedAt: ${preview.firstAccessedAt.toString()} (editorHelper, updateExercisePreviews)`)
        Logger.error(`hasLimitedWorkingTime: ${preview.hasLimitedWorkingTime} (editorHelper, updateExercisePreviews)`)
        Logger.error(
          `because of the last error we give the user the full working time... if that was a mistake the server should throw an error`)

        errorDialog(translate(TranslateIds.error),
          translate(TranslateIds.somethingWentWrongCalculatingLeftWorkingTime)
        )

        // firstAccessedAt is invalid --> DateHelper.getDiffInSeconds(firstAccessedAt, now) is NaN
        leftTimeString = preview.availableWorkingTimeInMinutes + ' min'
        // noinspection TsLint
        leftTimeInS = preview.availableWorkingTimeInMinutes * 60

        const copy: ExercisePreviewsFrontendExclusive = {
          ...preview,
          leftTimeString,
          leftTimeInS,
          isStillOpen: true
        }
        return copy

      } else {

        // noinspection TsLint
        const minutes = Math.floor(minDeadlineInS / 60)
        // noinspection TsLint
        const seconds = Math.floor(minDeadlineInS % 60)
        const result = minutes + ':' + padLeft(seconds, 2)
        leftTimeString = result
        leftTimeInS = minDeadlineInS

        if (minDeadlineInS <= 0) {
          //after the time is up we do not need to check for this any longer

          const copy: ExercisePreviewsFrontendExclusive = {
            ...preview,
            leftTimeString: translate(TranslateIds.timeUsedUp),
            leftTimeInS: 0,
            isStillOpen: false,
          }
          return copy

          //clearInterval(clearHandle)
        }

        const copy: ExercisePreviewsFrontendExclusive = {
          ...preview,
          leftTimeString,
          leftTimeInS,
          isStillOpen: minDeadlineInS > 0,
        }
        return copy
      }

    })

  return copies
}

function _userIsAllowedToSolveExercisePreview(preview: ExercisePreviewsFrontendExclusive, serverNow: Moment): boolean {

  if (!preview.isReleased) return false

  if (preview.isReleased && !preview.hasLimitedWorkingTime && preview.releaseStartType === ReleaseStartType.Manual) return true

  //TODO DUPLICATE see above
  if (preview.firstAccessedAt || preview.automaticEndAt || preview.hasLimitedWorkingTime) {

    // noinspection TsLint
    const availableWorkingTimeInSeconds: number = preview.availableWorkingTimeInMinutes * 60
    // noinspection TsLint
    let diffBeforeReleaseEnd = Number.POSITIVE_INFINITY
    // noinspection TsLint
    let diffToWorkingTimeDeadlineInSeconds = Number.POSITIVE_INFINITY

    if (preview.firstAccessedAt && preview.hasLimitedWorkingTime) {
      const diffInSeconds = serverNow.diff(preview.firstAccessedAt, 'seconds')
      diffToWorkingTimeDeadlineInSeconds = Math.floor(availableWorkingTimeInSeconds - diffInSeconds)
    }

    if (preview.automaticEndAt && preview.releaseStartType === ReleaseStartType.Automatic
      //&& preview.automaticEndAt.isValid()
      && preview.releaseDurationType === ReleaseDurationType.Limited) { //ReleaseDurationType.Limited --> preview.hasLimitedWorkingTime
      diffBeforeReleaseEnd = preview.automaticEndAt.diff(serverNow, 'seconds')
    }

    if (preview.automaticEndAt && preview.releaseStartType === ReleaseStartType.Automatic
      //&& preview.automaticEndAt.isValid()
      && preview.releaseDurationType === ReleaseDurationType.Infinitely) {
      diffBeforeReleaseEnd = preview.automaticEndAt.diff(serverNow, 'seconds')
    }

    const minDeadline = Math.min(diffToWorkingTimeDeadlineInSeconds, diffBeforeReleaseEnd)


    if (minDeadline === Number.POSITIVE_INFINITY) {

      Logger.error(
        "minDeadline was invalid (Number.POSITIVE_INFINITY) (editorHelper, _userIsAllowedToSolveExercisePreview)")
      Logger.error(
        `firstAccessedAt: ${preview.firstAccessedAt.toString()} (editorHelper, _userIsAllowedToSolveExercisePreview)`)
      Logger.error(
        `hasLimitedWorkingTime: ${preview.hasLimitedWorkingTime} (editorHelper, _userIsAllowedToSolveExercisePreview)`)

      errorDialog(translate(TranslateIds.error),
        translate(TranslateIds.somethingWentWrongCalculatingLeftWorkingTime)
      )

    } else if (isNaN(minDeadline)) {

      // firstAccessedAt is invalid --> DateHelper.getDiffInSeconds(firstAccessedAt, now) is NaN
      //assume user has the full working time

      Logger.error(
        "minDeadline was NaN (Number.POSITIVE_INFINITY) (editorHelper, _userIsAllowedToSolveExercisePreview)")
      Logger.error(
        `firstAccessedAt: ${preview.firstAccessedAt.toString()} (editorHelper, _userIsAllowedToSolveExercisePreview)`)
      Logger.error(
        `hasLimitedWorkingTime: ${preview.hasLimitedWorkingTime} (editorHelper, _userIsAllowedToSolveExercisePreview)`)
      Logger.error(
        `because of the last error we give the user the full working time... if that was a mistake the server should throw an error`)

      return true
    } else {

      if (minDeadline <= 0) {
        return false
      }
      //else still active/ open
    }
  }

  return true
}


/**
 * this is used so that the editor's componentWillUpdate method will be called with the right ForceEditorUpdateSessionDummy values
 * so that we get always a valid editor state (with undo, selection...)
 *
 *
 * if you wan't to call these funcs on your own see type MultiTabEditorState for usage
 *
 * @param {ReadonlyArray<SolutionFileDoExerciseFullBase>} solutionFiles
 * @param {(files: ReadonlyArray<SolutionFileDoExerciseFullBase>) => void} setSolutionFilesAction
 * @param {(state: MultiTabEditorState) => void} setForceEditorUpdateSessionDummy
 * @returns {AwaitActions}
 */
export function setSolutionFiles_AndRefreshEditor(solutionFiles: ReadonlyArray<ExerciseCodeTemplateFileFullBase>,
                                                  setSolutionFilesAction: (files: ReadonlyArray<ExerciseCodeTemplateFileFullBase>) => void,
                                                  setForceEditorUpdateSessionDummy: (state: MultiTabEditorState) => void): AwaitActions {
  return async (dispatch) => {

    dispatch(setForceEditorUpdateSessionDummy(MultiTabEditorState.beforeChangingFiles))

    dispatch(setSolutionFilesAction(solutionFiles))

    //if we don't wait then the editor won't re render and the change of MultiTabEditorState would not be noticed by
    //the editor...
    //TODO somehow make this better!!!

    // noinspection TsLint
    await refreshEditorDelay()

    dispatch(setForceEditorUpdateSessionDummy(MultiTabEditorState.afterChangingFiles))
    dispatch(setForceEditorUpdateSessionDummy(MultiTabEditorState.normal))
  }
}

//TODO see above!
export async function refreshEditorDelay(): Promise<void> {
  await wait(500)
}