/* tslint:disable */
/**
 * Created by janis dähne.
 */

import {SiteViewState} from "../types/siteView";
import Logger from "./logger";
import {SolutionVersionDoExerciseFullBase} from "../types/exerciseSolution";
import throttle from 'lodash-es/throttle'
import {dateTimeReviverLocalTime} from "./jsonTransformer";
import {AssessmentFullBase} from "../types/submissions";
import {SortDirection} from "../types/sortDirection";

//settings that are only stored at frontend

//1s should be ok, this is only that we save at MOST every 1s
//the save function is only triggered if something has changed
const throttleTimeInMs = 2000


interface SortingSettingTuple<T> {
  readonly sortByKey: keyof T
  readonly sortDirection: SortDirection
}

/**
 * use this as readonly and only change it through FrontendSettingsManager !!
 */
type State = {

  //--- site view states

  allOpenExercisesSiteViewState: SiteViewState

  allClosedExercisesSiteViewState: SiteViewState

  ownExercisesSiteViewState: SiteViewState

  groupExercisesSiteViewState: SiteViewState

  groupsSiteViewState: SiteViewState

  singleGroupSiteViewState: SiteViewState

  //--- site tags filter displayed

  allOpenExercisesSiteTagsFilterIsDisplayed: boolean,

  allClosedExercisesSiteTagsFilterIsDisplayed: boolean,

  ownExercisesSiteTagsFilterIsDisplayed: boolean,

  groupExercisesSiteTagsFilterIsDisplayed: boolean,


  //--- store sortings

  settingsSiteSorting: SortingSettingTuple<AssessmentFullBase> | null

  /**
   * the do editor/exercise release id (only overwrite files from backend if frontend has newer ones (some crash)
   * AND it must be the same release id (else we would overwrite the wrong files...)
   *
   * the last known one
   */
  lastKnownDoEditorReleaseId: number

  /**
   * the current exercise releases that the user solves
   */
  doEditorReleaseId: number

  /**
   * do editor files that should be saved (e.g. in case of a crash)
   *
   */
  doEditorSolution: SolutionVersionDoExerciseFullBase | null

  /**
   * a frontend setting is only valid for one user
   * if the user changes destroy it
   */
  lastLoggedInUserId: number


  /**
   * true: use (js) diff to display wrong chars
   * false: not (only display actual output read
   */
  useDiffIfOutputMismatched: boolean

  /**
   * true: replace whitespaces in console input and output with bullets (·)
   * false: not replace whitespaces
   */
  useReplaceWhiteSpacesWithBullets: boolean
}

const initial: State = {
  allOpenExercisesSiteViewState: SiteViewState.List,
  allClosedExercisesSiteViewState: SiteViewState.List,
  ownExercisesSiteViewState: SiteViewState.List,
  groupExercisesSiteViewState: SiteViewState.List,
  groupsSiteViewState: SiteViewState.List,
  singleGroupSiteViewState: SiteViewState.List,

  settingsSiteSorting: null,

  allOpenExercisesSiteTagsFilterIsDisplayed: false,
  allClosedExercisesSiteTagsFilterIsDisplayed: false,
  ownExercisesSiteTagsFilterIsDisplayed: false,
  groupExercisesSiteTagsFilterIsDisplayed: false,

  lastLoggedInUserId: -1,
  lastKnownDoEditorReleaseId: -1,
  doEditorReleaseId: -1,
  doEditorSolution: null,
  useDiffIfOutputMismatched: false,
  useReplaceWhiteSpacesWithBullets: true,
}

// noinspection TsLint
/**
 * represents the frontend settings that can be accessed globally
 *
 * the setters are normally called through the reducers
 * and in the reducers when we call reset we get the values from here (to overwrite reducer local initial values)
 *
 * because there is no explicit action call the frontend settings are transparent ... we can remove it and woul'd need
 * to change anything (except the calls to this class)
 *
 * TODO but maybe create a new reducer because we would load the settings from here in every reducer...
 * so we would get a lot of duplicate code...
 *
 * @type {State}
 */
let frontendSettings = initial

export class FrontendSettingsManager {

  static readonly localStorageKey = "frontendSettings"
  static hasReadFrontendSettingsFromLocalStorage = false

  static maxByteLocalStorage = 4194304 //should be 4 mb?

  /**
   * sometimes some components set properties earlier than expected (e.g. multi tab editor set solution after
   * creating the session and thus changing the solution before we can check for a new frontend version)
   *
   * so lock the settings until we are ready to take changes
   * @type {boolean}
   */
  static isFrozen = false


  static setIsFrozen(isFrozen: boolean): void {
    Logger.debug('frontendSettingsManager', `setting isFrozen to ${isFrozen}`)
    FrontendSettingsManager.isFrozen = isFrozen
  }

  static getIsFrozen(): boolean {
    if (FrontendSettingsManager.isFrozen) {
      Logger.debug('frontendSettingsManager', `settings are frozen!`)
    }

    return FrontendSettingsManager.isFrozen
  }

  //--- view states

  static setAllOpenExercisesSiteViewState(viewState: SiteViewState): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.allOpenExercisesSiteViewState = viewState
    saveFrontendSettingsThrottled()
  }

  static setAllClosedExercisesSiteViewState(viewState: SiteViewState): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.allClosedExercisesSiteViewState = viewState
    saveFrontendSettingsThrottled()
  }

  static setOwnExercisesSiteViewState(viewState: SiteViewState): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.ownExercisesSiteViewState = viewState
    saveFrontendSettingsThrottled()
  }

  static setGroupExercisesSiteViewState(viewState: SiteViewState): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.groupExercisesSiteViewState = viewState
    saveFrontendSettingsThrottled()
  }

  static setGroupsSiteViewState(viewState: SiteViewState): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.groupsSiteViewState = viewState
    saveFrontendSettingsThrottled()
  }

  static setSingleGroupSiteViewState(viewState: SiteViewState): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.singleGroupSiteViewState = viewState
    saveFrontendSettingsThrottled()
  }

  static setSubmissionsSiteSorting(sorting: SortingSettingTuple<AssessmentFullBase>): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.settingsSiteSorting = sorting
    saveFrontendSettingsThrottled()
  }

  //--- is tags filter displayed

  static setAllOpenExercisesSiteTagsFilterIsDisplayed(isDisplayed: boolean): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.allOpenExercisesSiteTagsFilterIsDisplayed = isDisplayed
    saveFrontendSettingsThrottled()
  }

  static setAllClosedExercisesSiteTagsFilterIsDisplayed(isDisplayed: boolean): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.allClosedExercisesSiteTagsFilterIsDisplayed = isDisplayed
    saveFrontendSettingsThrottled()
  }

  static setOwnExercisesSiteTagsFilterIsDisplayed(isDisplayed: boolean): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.ownExercisesSiteTagsFilterIsDisplayed = isDisplayed
    saveFrontendSettingsThrottled()
  }

  static setGroupExercisesSiteTagsFilterIsDisplayed(isDisplayed: boolean): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.groupExercisesSiteTagsFilterIsDisplayed = isDisplayed
    saveFrontendSettingsThrottled()
  }


  //--- other
  static setUseDiffIfOutputMismatched(useDiffIfOutputMismatched: boolean): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.useDiffIfOutputMismatched = useDiffIfOutputMismatched
    saveFrontendSettingsThrottled()
  }

  static setUseReplaceWhiteSpacesWithBullets(useReplaceWhiteSpacesWithBullets: boolean): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.useReplaceWhiteSpacesWithBullets = useReplaceWhiteSpacesWithBullets
    saveFrontendSettingsThrottled()
  }

  /**
   * ONLY PASS DOWN A COPY! else we could change the original reference
   * @param {SolutionVersionDoExerciseFullBase} doEditorSolution
   */
  static setDoEditorSolution(doEditorSolution: SolutionVersionDoExerciseFullBase | null): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.doEditorSolution = doEditorSolution
    Logger.log("frontendSettingsManager", "saving frontend solution ... (but throttled")
    saveFrontendSettingsThrottled()
  }

  static setLastKnownDoEditorReleaseId(lastKnownDoEditorReleaseId: number): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.lastKnownDoEditorReleaseId = lastKnownDoEditorReleaseId
    saveFrontendSettingsThrottled()
  }

  static setDoEditorReleaseId(doEditorReleaseId: number): void {

    if (FrontendSettingsManager.getIsFrozen()) return

    frontendSettings.doEditorReleaseId = doEditorReleaseId
    saveFrontendSettingsThrottled()
  }

  static setLastLoggedInUserId(lastLoggedInUserId: number): void {

    //if (FrontendSettingsManager.getIsFrozen()) return //always allow this

    frontendSettings.lastLoggedInUserId = lastLoggedInUserId
    saveFrontendSettingsThrottled()
  }



  /**
   * when a new user logs in (user changed) or the exercise changed and thus the editor content is wrong/old
   * clear the storage
   */
  static clear(): void {
    frontendSettings = {...initial}
    forceDestroy();
  }

  static getFrontendSettings(): State {
    if (FrontendSettingsManager.hasReadFrontendSettingsFromLocalStorage === false) {
      FrontendSettingsManager.hasReadFrontendSettingsFromLocalStorage = true
      if (supportsLocalStorage()) {
        Logger.debug("frontendSettingsManager", "reading frontend settings")
        const json = localStorage.getItem(FrontendSettingsManager.localStorageKey)

        if (json) {
          try {
            frontendSettings = FrontendSettingsManager.setFrontendSettingsFromJsonString(json)
          } catch (err) {
            Logger.debug("frontendSettingsManager", "invalid settings")
          }

        } else {
          Logger.debug("frontendSettingsManager", "no settings found")
        }
      }
    }

    return frontendSettings
  }

  static forceSaveFrontendSettings(): void {
    forceSave();
  }

  static getFrontendSettingsAsJsonString(): string {

    //some replacements of props we don't need
    return JSON.stringify(frontendSettings, (key: string, value) => {

      if (key === 'uiEditSession'){
        return ''
      }
      if (key === 'shouldNotCount') {
        return ''
      }
      return value
    })
  }

  static setFrontendSettingsFromJsonString(json: string): State {
    //use same as in setup.ts to convert entities from backend (dateTimeReviverLocalTime converts dates to moment objs)
    const maybeSettings = JSON.parse(json, dateTimeReviverLocalTime)

    // if (maybeSettings && objectsHaveSameKeys(frontendSettings, maybeSettings)) {
    //   //TODO more checks ... also value types & ranges??
    //   //or just take valid keys (with values)
    //   frontendSettings = maybeSettings
    //   Logger.debug("frontendSettingsManager", "read frontend settings were valid (at least same keys)")
    // } else {
    //   Logger.warn( "[frontendSettingsManager] found invalid/old settings [not the same object keys]")
    // }

    //we now only take values that we know (e.g. if we create a property)
    if (maybeSettings) {
      for (const key in frontendSettings) {

        if (maybeSettings.hasOwnProperty(key)) {
          //take value
          frontendSettings[key] = maybeSettings[key]
        }
        else {
          Logger.warn(` [frontendSettingsManager] settings key not found in old settings: ${key}, skipping`)
        }
      }
    }


    return frontendSettings
  }
}

/**
 * func to save the frontend settings but throttled
 * @type {() => void}
 */
const saveFrontendSettingsThrottled: () => void = _saveFrontendSettingsThrottled()


//from https://stackoverflow.com/questions/14368596/how-can-i-check-that-two-objects-have-the-same-set-of-property-names
// noinspection TsLint
function objectsHaveSameKeys(...objects: any[]): boolean {
  const allKeys = objects.reduce((keys, object) => keys.concat(Object.keys(object)), []);
  const union = new Set(allKeys);
  return objects.every(object => union.size === Object.keys(object).length);
}

function _saveFrontendSettingsThrottled(): () => void {
  // noinspection TsLint
  return throttle(() => {
    forceSave()
  }, throttleTimeInMs)
}

function forceDestroy(): void {
  if (supportsLocalStorage()) {
    localStorage.removeItem(FrontendSettingsManager.localStorageKey)
  }
}

function forceSave(): void {
  if (supportsLocalStorage()) {
    Logger.debug("frontendSettingsManager", "writing frontend settings")

    const json = FrontendSettingsManager.getFrontendSettingsAsJsonString()
    if (byteLength(json) >= FrontendSettingsManager.maxByteLocalStorage) {
      Logger.warn(
        'some browsers only support 5 MByte local storage ... the current settings are larger than 4 MByte! So the save might not work...')
    } else {
      //ok
    }
    localStorage.setItem(FrontendSettingsManager.localStorageKey, json)
  }
}

//from https://www.w3schools.com/html/html5_webstorage.asp
function supportsLocalStorage(): boolean {
  if (typeof(Storage) !== "undefined") {
    // Code for localStorage/sessionStorage.
    return true
  } else {
    // Sorry! No Web Storage support..
    return false
  }
}

//from https://stackoverflow.com/questions/5515869/string-length-in-bytes-in-javascript
// noinspection TsLint
function byteLength(str: string): number {
  // returns the byte length of an utf8 string
  // noinspection TsLint
  let s = str.length;
  // noinspection TsLint
  for (let i = str.length - 1; i >= 0; i--) {
    // noinspection TsLint
    let code = str.charCodeAt(i);
    // noinspection TsLint
    if (code > 0x7f && code <= 0x7ff) s++;
    // noinspection TsLint
    else { // noinspection TsLint
      if (code > 0x7ff && code <= 0xffff) s += 2;
    }
    // noinspection TsLint
    if (code >= 0xDC00 && code <= 0xDFFF) i--; //trail surrogate
  }
  return s;
}
