/**
 * Created by janis dähne.
 */

import axios, {AxiosResponse} from 'axios'
import {dateTimeReviverLocalTime} from "./helpers/jsonTransformer";
import Logger from "./helpers/logger";

//we include this via html
//import '../node_modules/semantic-ui-css/semantic.min.css';

//import marked = require("marked")
import marked = require('../externalLibsForBundling/markedChanged/marked')

require('react-datetime/css/react-datetime.css')
//require('../customChangedLib/react-notifications/notifications.styl') //imported via allStylues.styl

//TODO https://bjacobel.com/2016/12/04/highlight-bundle-size/
//hightlight js
require('highlight.js/styles/atom-one-light.css')
require('github-markdown-css/github-markdown.css')

require('sweetalert2/dist/sweetalert2.min.css')

//import before... to be sure??? see https://craig.is/killing/mice --> plugins
import * as mousetrap from 'mousetrap'

require('mousetrap/plugins/global-bind/mousetrap-global-bind.js')
require('mousetrap/plugins/pause/mousetrap-pause.js')

import store from './state/state'
import {setUserData} from "./state/actions/settingsSite/userSettingsSiteActions";
import {userDataInitial} from "./state/reducers/settingsSite/settingsSiteReducers";
import {escapeHtml, insertLineNumbers} from "./helpers/stringHelper";
import {MarkedOptions} from "marked";
import {showMarkdownCodeLineNumbers} from "./constants";
import {ServerResponse} from "./types/response";
import {ResponseCodes} from "./helpers/responseCodes";
import {errorDialog} from "./helpers/dialogHelper";
import {translate} from "../i18n/translationHelper";
import {TranslateIds} from "../i18n/ids";
import {isProduction} from "../debug";
const hljs =  require('highlight.js/lib/highlight')

//https://bjacobel.com/2016/12/04/highlight-bundle-size/
//import * as hljs from 'highlight.js';

const javascript = require('highlight.js/lib/languages/javascript')
const java = require('highlight.js/lib/languages/java')
const cpp = require('highlight.js/lib/languages/cpp')
const diff = require('highlight.js/lib/languages/diff')
const cs = require('highlight.js/lib/languages/cs') //c#
const sql = require('highlight.js/lib/languages/sql')
const json = require('highlight.js/lib/languages/json')

hljs.registerLanguage('javascript', javascript)
hljs.registerLanguage('java', java)
hljs.registerLanguage('cpp', cpp)
hljs.registerLanguage('diff', diff)
hljs.registerLanguage('cs', cs)
hljs.registerLanguage('sql', sql)
hljs.registerLanguage('json', json)

/**
 * various setup steps for the app
 *
 * this should be called before everything else
 */
export function setup(): void {

  Logger.log('setup started...')

  //https://developer.mozilla.org/de/docs/Web/API/GlobalEventHandlers/onerror
  window.onerror = function (message:any, file: string, line: number, col: number, error: Error): boolean {
    Logger.warn(`catched unhandled error!! ${JSON.stringify(message)};${file};${line};${col};${JSON.stringify(error)}`)
    return false //throw error
  };

  const markedRenderer = new marked.Renderer()

  const markedOptions: MarkedOptions = {
    renderer: markedRenderer,
    gfm: true,
    tables: true,
    breaks: true,
    pedantic: false,
    sanitize: false,
    smartLists: true,
    smartypants: false,
    highlight: (code: any, lang: any) => {
      if (lang === undefined) return code
      return hljs.highlight(lang, code).value
    },
    langPrefix: 'lang-', //from marked
  }

  marked.setOptions(markedOptions)

  if (showMarkdownCodeLineNumbers) {
    const spacesAfterLineNumber = '   '
    //copied from marked, then modified
    markedRenderer.code = (code: string, lang: string, escaped: boolean) => {
      //split into lines and add line numbers
      if (markedOptions.highlight) {
        const out = markedOptions.highlight(code, lang);
        if (out !== null && out !== code) {
          escaped = true;
          code = insertLineNumbers(out, spacesAfterLineNumber);
        } else {
          code = code
        }
      } else {
        code = code
      }

      if (!lang) {
        return '<pre><code>'
          + (escaped ? code : escapeHtml(code, true))
          + '\n</code></pre>';
      }

      return '<pre><code class="'
        + markedOptions.langPrefix
        + escapeHtml(lang, true)
        + '">'
        + (escaped ? code : escapeHtml(code, true))
        + '\n</code></pre>\n';

    }
  }

  //axios.defaults.responseType = 'json'

  //https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/withCredentials
  //we want cross-origin requests to include cookies for develop...
  if (!isProduction) {
    axios.defaults.withCredentials = true
  }

  axios.defaults.transformResponse = (response: string) => {
    //this removes the default json transformer and we will get the plain string
    return response
  }

  // axios.defaults.headers = {
  //   ...axios.defaults.headers,
  //   post: {
  //     'content-type': 'application/json'
  //   }
  // }

  //logged in? interceptor
  axios.interceptors.response.use((response) => {
      return response
    },
    (error: { readonly response: AxiosResponse }) => {

      if (!error.response) {
        return error
      }

      // noinspection TsLint
      if (error.response.status === 401) {
        //reset user data to show login
        Logger.log("unauthorized... resetting user data to show login")
        store.dispatch(setUserData(userDataInitial))
      }
      return Promise.reject(error.response)
    }
  )



  axios.interceptors.response.use((response: AxiosResponse) => {


    if (!response.headers['Content-Type'] || response.headers['Content-Type'] === 'application/json') {
      if (response.data) {
        //this will convert all string dates into moment objects
        const obj: ServerResponse = JSON.parse(response.data, dateTimeReviverLocalTime)

        //server errors are normal json objects with some special codes

        if (obj.code !== ResponseCodes.ok) {
          //some error/warning
          Logger.error(`setup catched not OK response code: ${obj.code}, message: ${obj.message}` )
          errorDialog(translate(TranslateIds.serverError), obj.message)
          //reject
          //return Promise.reject(error.response)
        }

        response.data = obj
      }
    }
    return response
  })


  Logger.log('...setup finished')


}
