/**
 * Created by janis dähne.
 */

/*
  xxxFullBase (without referring xxxForBackend and xxxFromBackend)
  are types that are used for sending to backend and receiving from backend

  else we have xxxForBackend for sending to backend and
  xxxFromBackend for receiving from backend and
  xxxBase for common properties
 */

import {TestTypeFullBase} from "./testTypes";
import {TestAnswerFromBackend} from "./testsAndTestProtocol";
import {AceEditSession} from "../components/codeEditors/multiTabCodeEditorProps";

/**
 * the exercise base type (from and for backend) normally for the exercise editor
 */
export interface EditorExerciseBase {
  readonly id: number
  /**
   * the exercise properties
   */
  readonly exerciseProperties: EditorExercisePropertiesFullBase
  /**
   * the exercise description with the assets
   */
  readonly exerciseDescription: ExerciseDescriptionFullBase
  /**
   * the exercise meta data (currently only tags)
   */
  readonly metaData: EditorExerciseMetaDataFullBase

  /**
   * the default test case settings, used when the user creates a custom test
   */
  readonly defaultTestSettings: TestSettingsFullBase
  /**
   * the code templates for the plangs
   */
  readonly codeTemplates: ReadonlyArray<ExerciseCodeTemplateFullBase>

  /**
   * some notes
   */
  readonly note: string
}
/**
 * a exercise from the frontend for the backend (tests does include file content)(create/update exercise)
 */
export interface EditorExerciseExerciseForBackend extends EditorExerciseBase{

  /**
   * the user id of the creator or null
   * THIS IS ONLY THERE SO THAT THE STORE CAN USE this type instead of xxxFromBackend
   * so we convert EditorExerciseFromBackend to this type
   */
  readonly creatorId?: number

  /**
   * the tests
   */
  readonly tests: ReadonlyArray<ExerciseTestForBackend>
}

/**
 * a exercise from the frontend for the backend (tests does NOT include file content because is cannot changed or viewed in the frontend...)
 */
export interface EditorExerciseFromBackend extends EditorExerciseBase {

  /**
   * the user id of the creator or null (if the owner was deleted, then only the user group can manage the exercise)
   */
  readonly creatorId?: number
  /**
   * the tests
   */
  readonly tests: ReadonlyArray<ExerciseTestFromBackend>
}

/**
 * the properties for an exercise (in the backend these are part of the exercise but not in the frontend (have own panel)
 * --> no id
 */
export interface EditorExercisePropertiesFullBase {
  /**
   * the display name of the exercise
   */
  readonly displayName: string
  /**
   * the short description
   */
  readonly shortDescription: string
  /**
   * true: displays the submit results to the students, false: not
   */
  readonly showSubmitResults: boolean

  /**
   * true: user can create new files, false: can only modify the given files
   */
  readonly canUserCreateFiles: boolean
  /**
   * true: user can create custom tests, false: not
   */
  readonly canUserCreateCustomTests: boolean
  /**
   * true: the exercise is only visible to the owner of the exercise and only the owner can do something with the exercise (update, release,...) false: not
   */
  readonly isOnlyVisibleToMe: boolean
  /**
   * the group id where the exercise will belong to
   */
  readonly userGroupId: number
}

/**
 * the meta data for the exercise (currently only tags)
 */
export interface EditorExerciseMetaDataFullBase {
  readonly id: number
  /**
   * the list of tag ids
   */
  readonly tagIds: ReadonlyArray<number>
}


/**
 * the exercise description common type
 */
export interface ExerciseDescriptionFullBase {
  readonly id: number
  /**
   * the content
   */
  readonly content: string
  /**
   * the assets (files)
   */
  readonly assets: ReadonlyArray<ExerciseAssetFullBase>
}

/**
 * the exercise asset for the exercise description, because we need the content to display e.g. an image...
 */
export interface ExerciseAssetFullBase {
  readonly id: number
  /**
   * the display name
   */
  readonly displayName: string
  /**
   * the asset url (normally asset://...), used to includ the asset in the exercise description
   * so the user can type 'user friendly' name and does not have to use the browser blob url
   */
  readonly assetUrl: string
  /**
   * the mime type
   */
  readonly mimeType: string
  /**
   * the content
   *  backend --> frontend its a list | frontend --> backend its an array... (maybe use same approach as testing...
   *  TODO MAYBE when using byte[] in frontend then its send as base64 encoded ??
   */
  readonly content: ReadonlyArray<number>
}

/**
 * an exercise code template for a plang
 */
export interface ExerciseCodeTemplateFullBase {
  readonly id: number

  /**
   * the plang this code template is for
   */
  readonly pLangId: number

  /**
   * the main file
   */
  readonly mainFileId: number

  /**
   * the code template files
   */
  readonly templateFiles: ReadonlyArray<ExerciseCodeTemplateFileFullBase>

}

/**
 * a code template file, some lines of code the tutor wants to give the students (assistance)
 */
export interface ExerciseCodeTemplateFileFullBase {
  readonly id: number
  /**
   * the file name with extension
   */
  readonly fileNameWithExtension: string
  /**
   * the order index how the file is displayed in the editor
   * NOT USED YET
   * -x is not displayed (only in explorer)
   */
  readonly displayIndex: number
  /**
   * true: user can edit the file, false: not
   */
  readonly isEditableByUser: boolean
  /**
   * the content of the file
   */
  readonly content: string

  /**
   * ui property
   * only set through editor (multi tab editor)... (not in reducers from backend type
   *
   * the ace edit session so that we reconstruct all sessions & states from a blank view (from scratch) (e.g. because login was displayed)
   */
  readonly uiEditSession: AceEditSession | null | undefined

}


/**
 * a test base type
 */
export interface ExerciseTestBase {
  readonly id: number
  /**
   * the display name
   */
  readonly displayName: string
  /**
   * true: is a submit test, false: not (normal test)
   */
  readonly isSubmitTest: boolean
  /**
   * the test content
   */
  readonly content: string
  /**
   * the weight of the test
   */
  readonly weight: number
  /**
   * ui property, the display index for the sorting in the frontend
   * (>= 0, 0 is first, 1 second, ...)
   */
  readonly displayIndex: number

  /**
   * the test type id
   */
  readonly testTypeId: number

  /**
   * the test settings
   */
  readonly testSettings: TestSettingsFullBase
}

/**
 * a test for the backend (assets does contain data)
 */
export interface ExerciseTestForBackend extends ExerciseTestBase {
  /*
   * the assets (files) (with data)
   */
  readonly files: ReadonlyArray<TestAssetForBackend>
}
/**
 * a test from the frontend (assets does NOT contain data)
 */
export interface ExerciseTestFromBackend extends ExerciseTestBase {

  /**
   * the assets (files) (without data because could not viewed or changed at frontend)
   */
  readonly files: ReadonlyArray<TestAssetFromBackendWithData>
}

/**
 * a runnable test (with result...)
 */
export interface RunnableExerciseTestFrontendExclusive extends ExerciseTestFromBackend {
  /**
   * true: test is running, false: not
   * used to display a spinner
   */
  readonly uiIsRunning: boolean
  /**
   * the result for this test
   *
   */
  readonly result: TestAnswerFromBackend | null
}


/**
 * test settings for a test
 */
export interface TestSettingsFullBase {
  readonly id: number
  /**
   * the test timeout in ms
   */
  readonly timeoutInMs: number
  /**
   * the memory limit in kb
   */
  readonly memoryLimitInKb: number

  /**
   * the max disk space to users program can write to
   */
  readonly maxDiskSpaceInKb: number
}

/**
 * a test asset
 */
export interface TestAssetBase {
  readonly id: number,
  /**
   * the display name
   */
  readonly displayName: string
  /**
   * the mime type
   */
  readonly mimeType: string
  /**
   * the size of the asset (when sending to frontend do not include content)
   * because could not be changed or viewed
   */
  readonly size: number
}

/**
 * a test asset for the backend (includes data/content)
 */
export interface TestAssetForBackend extends TestAssetBase {

  /**
   * the content of the asset
   */
  readonly content: ReadonlyArray<number>
}

/**
 * TODO rename
 * the name is misleading ... the content can still be exlucded by the backend
 */
export interface TestAssetFromBackendWithData extends TestAssetBase {
  /**
   * the content of the asset
   * CAN BE NULL/not set if the backend does not include the content
   */
  readonly content: ReadonlyArray<number> | null
}


/**
 * an annotation for the editor
 */
export interface TestEditorAnnotation {
  /**
   * the row
   */
  readonly row: number
  /**
   * the column
   */
  readonly column: number
  /**
   * the text to display
   */
  readonly text: string
  /**
   * the type (currently ace editor types)
   */
  readonly type: string | 'error' | 'warning' | 'info'
}

/**
 * same as TestEditorAnnotation with a file name
 */
export interface TestEditorAnnotationWithFileName extends TestEditorAnnotation {
  /**
   * the file name for the annotation
   */
  readonly fileNameWithExtension: string
}
