/**
 * Created by janis dähne.
 */


import {Moment} from "moment";
import {ReleaseDurationType} from "./ReleaseDurationType";
import {ReleaseStartType} from "./ReleaseStartType";


/**
 * represents a base exercise release
 */
export interface ExerciseReleaseBase {
  readonly id: number
  /**
   * true: is released, false: not
   */
  readonly isReleased: boolean
  /**
   * the generated code
   */
  readonly generatedCode: string

  /**
   * the available working time in minutes if limitedWorkingTime is true
   *
   * can be null if invalid input (that does not mean there is no working time)
   * only the input is invalid (this should be checked by the frontend)
   */
  readonly availableWorkingTimeInMinutes: number | null
  /**
   * true: is visible to all after the release
   */
  readonly isVisibleToAllAfterRelease: boolean
  /**
   * true: unlimited time to solve the exercise, false: not --> see availableWorkingTimeInMinutes
   */
  readonly hasLimitedWorkingTime: boolean
  /**
   * a note
   */
  readonly note: string
  /**
   * true: run the automatic assessment, false: not
   */
  readonly shouldAutomaticAssessSubmissions: boolean

  /**
   * true: submit tests and normal tests are run, false: only submit tests are run
   */
  readonly runAlsoNormalTests: boolean

  /**
   * true: automatic assessment has started, false: not
   */
  readonly hasAutomaticAssessmentStarted: boolean

  /**
   * true: automatic assessment has finished, false: not
   */
  readonly hasAutomaticAssessmentFinished: boolean

  /**
   * true: some error occurred in the last automatic assessment (after the first error all other assessments are skipped),
   * false: everything ok
   */
  readonly hadAutomaticAssessmentErrors: boolean

  /**
   * the plang to user can use or null then the user can solve the exercise in any plang (where a template is provided)
   *
   * with yapex 1.0 we no longer support any plang ... the user must select one
   * (any plang is useless because normally we want X plangs and not all)
   *
   * for
   */
  readonly pLangId: number | null

  /**
   * Infinitely = 0, Limited = 1
   */
  readonly releaseDurationType: ReleaseDurationType
  /**
   * Manual = 0, Automatic = 1
   */
  readonly releaseStartType: ReleaseStartType

  /**
   * the date time when the release as actually released (usefully when manually released)
   *
   * only set when we got the release from server and it was released...
   */
  readonly releasedAt: Moment | null

  /**
   * e.g. for exams we don't want students to see
   * the exercises after the exam
   *
   * so ensure that if this is
   * true: the exercise is not displayed to the client (in open & closed exercises...)
   * false: a normal exercise/release
   */
  readonly hideInOverviews: boolean

  /**
   * e.g. in an electronic exam (e-exam) we might have multiple windows in e.g. the safe exam browser
   * the student could copy & past exercises from other systems into yapex and try to compile, execute code...
   *
   * true: try to clear the clipboard if the tab is changed...
   *     show an overlay every time the student tabs or clicks outside of the browser window
   *     and when he/she wants to resume he/she must click a button to proceed
   * false: normal
   *
   * frontend only option...
   */
  readonly shouldClearClipboard: boolean

  /**
   * in the do exercise site the link mode always overwrites this setting!
   *
   * true: hide the site header bar (where the menu, user menu is located)
   *    so that a student cannot navigate away from the do exercise site (e.g. in an exam)
   *
   * false: not hide the bar
   */
  readonly hideSiteHeaderBar: boolean
}

/**
 * the release receiving from the backend
 */
export interface ExerciseReleaseFromBackend extends ExerciseReleaseBase {
  /**
   * the automatic starting date time (if the release type is automatic)
   */
  readonly automaticStartAt: Moment | null
  /**
   * the automatic ending date time (if the release type is automatic)
   */
  readonly automaticEndAt: Moment | null

  /**
   * the amount of solutions in this release
   */
  readonly solutionsCount: number
}


/**
 * a tuple to create an exercise release
 */
export interface ExerciseReleaseCreateTupleForBackend {
  /**
   * the id of the exercise
   */
  readonly forExerciseId: number
  /**
   * the release
   */
  readonly release: ExerciseReleaseForBackend
}

/**
 * backend expects dates not strings
 */
export interface ExerciseReleaseForBackend extends ExerciseReleaseBase {
  /**
   * the automatic starting date time (if the release type is automatic)
   */
  readonly automaticStartAt?: Date
  /**
   * the automatic ending date time (if the release type is automatic)
   */
  readonly automaticEndAt?: Date
}
