/**
 * Created by janis dähne.
 */

import {SolutionFileDoExerciseFullBase} from "./exerciseSolution";
import {getPropertyName} from "../helpers/typeHelpers";
import endsWith from 'lodash-es/endsWith'

export interface RegexTestObject {
  readonly mode?: string
  readonly connector?: string

  //now 1 regex expression for 1 file
  // noinspection TsLint
  [fileName: string]: string
}

export interface RegexObj {
  readonly pattern: string
  readonly flags: string | null
}

export interface RegexTestObjectConverted {
  //1 regex expression can consists of multiple regex expressions
  //also remove the /regex/ here (the /)
  // noinspection TsLint
  [fileName: string]: RegexObj[]
}


export function parseRegexTestContent(content: string): RegexTestObject | string {

  const result: RegexTestObject = {}

  if (content.trim() === '') {
    return 'no content'
  }

  const lines = content.split('\n')

  if (lines.length === 0) {
    return 'no lines'
  }

  // noinspection TsLint
  for (let i = 0; i < lines.length; i++) {
    const line = lines[i];

    if (line.trim() === '') {
      continue
    }

    const colonPos = line.indexOf(':')

    if (colonPos === -1) {
      return `no : found, error on line ${i + 1}`
    }

    //this could be mode or connector too ... that's ok
    const fileName = line.substring(0, colonPos).trim()
    const regex = line.substring(colonPos + 1).trim()

    if (fileName === '') {
      return `no file name found on line ${i + 1}`
    }

    if (regex === '') {
      return `no regex found on line ${i + 1}`
    }

    result[fileName] = regex
  }

  return result

}

/**
 * converts and checks a RegexTestObject (only allowed mode, connector)
 * also removes the /.../ from the regex so checkRegexs can process it
 * @param {RegexTestObject} obj
 * @returns {RegexTestObjectConverted}
 */
export function convertRegexTestObject(obj: RegexTestObject): RegexTestObjectConverted | null {

  const keys = Object.keys(obj)

  const result: RegexTestObjectConverted = {}

  // noinspection TsLint
  let hasSomeError = false

  // noinspection TsLint
  for (let i = 0; i < keys.length; i++) {
    const key = keys[i]
    const value = obj[key]

    if (key === getPropertyName<RegexTestObject>('mode')) {
      if (value === 'blacklist' || value === 'whitelist') {
        //all ok
      } else {
        return null
      }
      continue
    }
    if (key === getPropertyName<RegexTestObject>('connector')) {
      if (value === '&&' || value === '||') {
        //all ok
      } else {
        return null
      }
      continue
    }
    const values = value.split(/&&|\|\|/)
    result[key] = values.map(p => {

      p = p.trim()

      if (p[0] !== '/' || p[p.length - 1] !== '/' || p.length === 1) { //if length 1 == 1 and the sting is / this would fail


        //we now support regex flags...
        //so check

        if (p[0] === '/' && endsWith(p, '/')) {
          return {
            pattern: p.substring(1, p.length - 1),
            flags: null
          }
        }
        // only first result matters
        // else if (p[0] !== '/' && endsWith(p, '/g')) {
        //   return {
        //     pattern: p.substring(1, p.length - 1),
        //     flags: 'g'
        //   }
        // }
        else if (p[0] === '/' && endsWith(p, '/m')) {
          return {
            pattern: p.substring(1, p.length - 1),
            flags: 'm'
          }
        }
        else if (p[0] === '/' && endsWith(p, '/i')) {
          return {
            pattern: p.substring(1, p.length - 1),
            flags: 'i'
          }
        }
        else if (p[0] === '/' && endsWith(p, '/im')) {
          return {
            pattern: p.substring(1, p.length - 1),
            flags: 'mi'
          }
        }
        else if (p[0] === '/' && endsWith(p, '/mi')) {
          return {
            pattern: p.substring(1, p.length - 1),
            flags: 'mi'
          }
        }

        hasSomeError = true

      }

      return {
        pattern: p.substring(1, p.length - 1),
        flags: null
      }
    })
  }

  if (hasSomeError) return null

  return result
}

/**
 * checks a regex tests object if all regex are valid
 * @param {RegexTestObjectConverted} obj
 * @returns {ReadonlyArray<string>}
 */
export function checkRegexs(obj: RegexTestObjectConverted): ReadonlyArray<string> {
  const keys = Object.keys(obj)

  // noinspection TsLint
  let results = []

  // noinspection TsLint
  for (let i = 0; i < keys.length; i++) {
    const key = keys[i]
    const values = obj[key]

    // noinspection TsLint
    for (let k = 0; k < values.length; k++) {
      try {

        const test = values[k].flags
          ? new RegExp(values[k].pattern, values[k].flags)
          : new RegExp(values[k].pattern)

      } catch (err) {
        //errors still include the /.../ because that are js regex... that's ok
        results.push(`${key} - ${err.message}`)
      }
    }
  }

  return results as ReadonlyArray<string>
}


export interface RunCompileSingleFileCommandBase {
  /**
   * the p lang id to use
   */
  readonly plangId: number
  /**
   * the file to use as main file
   */
  readonly mainFileName: string
  /**
   * the file to compile (syntax check)
   */
  readonly files: ReadonlyArray<SolutionFileForTestServer>
}

/**
 *  a command to compile a single file in order to validate the syntax
 *
 *  we send all files but set the main file to the given file name...
 */
export interface RunCompileSingleFileCommandForBackend extends RunCompileSingleFileCommandBase {
  /**
   * the code from the exercise release
   */
  readonly generatedCode: string
}


export interface JustRunProgramCommandForBackend {
  /**
   * the code from the exercise release
   */
  readonly releaseCode: string

  /**
   * the p lang id to use
   */
  readonly plangId: number

  /**
   * the args to use
   */
  readonly CmdArgs: string
}


export interface JustRunProgramTestCustomProjectCommandForBackend {

  readonly customProjectId: number
  /**
   * the p lang id to use
   */
  readonly plangId: number

  /**
   * the args to use
   */
  readonly CmdArgs: string
}

export interface RunCompileSingleFileCommandCustomProjectForBackend extends RunCompileSingleFileCommandBase {
  /**
   * the custom project (must be a project of the current user (so not every user can compile random)
   */
  readonly customProjectId: number
}

/**
 * a run test command with a solution to use
 */
export interface RunAfterTestCommandForBackend {
  readonly releaseCode: string
  readonly plangId: number
  readonly testIds: ReadonlyArray<number>
  readonly isNormalTest: boolean
  readonly isSubmitTest: boolean
  readonly mainFileId: number
  readonly solutionFiles: ReadonlyArray<SolutionFileDoExerciseFullBase>
}

/**
 * a file for the test server
 */
export interface SolutionFileForTestServer {
  /**
   * the file name WITH extension
   *
   * we can't use the file id because the file might not be saved to backend yet
   */
  readonly fileName: string
  /**
   * the file content
   */
  readonly fileContent: string
}

/**
 * a test request/command for backend
 */
export interface RunTestCommandForBackend {

  /**
   * the code from the exercise release
   */
  readonly generatedCode: string
  /**
   * the p lang id to use
   */
  readonly plangId: number
  /**
   * the test ids... also ensure that the test is not a submit test
   */
  readonly testIds: ReadonlyArray<number>

  /**
   * true: run a normal test (search for the test in the normal tests), false: not a normal test
   */
  readonly isNormalTest: boolean

  /**
   * true: run a custom test (search for the test in the custom tests), false: not a custom test
   */
  readonly isCustomTest: boolean

  /**
   * true: run a submit test (search for the test in the submit tests), false: not a submit test
   */
  readonly isSubmitTest: boolean

  /**
   * int: the solution version, null: to use the latest version
   */
  readonly solutionVersionId: number | null
}


export interface RunTestCommandCustomProjectForBackend {

  readonly customProjectId: number
  /**
   * the p lang id to use
   */
  readonly plangId: number

  readonly testIds: ReadonlyArray<number>
}

export interface TestAnswerFromBackend {

  /**
   * the code for the test (see response codes from test server)
   *
   * when coming from a stored test result this is empty
   */
  readonly testServerCode: number | null

  /**
   * some more information about the test if any error occurred (for debugging)
   *
   * when coming from a stored test result this is empty
   */
  readonly testServerMessage: string | null
  /**
   * the test id
   */
  readonly testId: number
  /**
   * true: test was successful, false: not
   */
  readonly passed: boolean
  /**
   * true: has compiled, fasle: not or if the test does not included a compile step this is also false
   */
  readonly hasCompiled: boolean
  /**
   * the exit code of the user program, not used yet (not event set)
   */
  readonly programExitCode: number
  /**
   * the test result code (see test server codes)
   */
  readonly testResultCode: number
  /**
   * the conversation protocol between the user program and the test runner
   */
  readonly protocol: ReadonlyArray<string>
}


/**
 * some constants for the test protocol
 *
 * the test protocol lines can start with these characters...
 */
export enum TestProtocol {
  /**
   * the test content string the starts the argument line
   */
  argumentSign = '$',
  /**
   * the test content string that indicates a output line (user program output)
   */
  outputSign = '>',
  /**
   * the test content string that indicates a input line (test server --> users program)
   */
  inputSign = '<',

  /**
   * the test line string that indicates a black box test regex line (output must match a regex)
   */
  regexOutputSign = 'r',

  /**
   * the protocol string that indicates an error
   */
  errorPrefix = 'error:',
  /**
   * the protocol string that indicates a output line that was expected but not outputted from the user program
   */
  expectedPrefix = 'expected:',

  /**
   * the protocol string that indicates the end of the test
   */
  endOfTestString = 'EOT'
}


